/**
 * PANDA 3D SOFTWARE
 * Copyright (c) Carnegie Mellon University.  All rights reserved.
 *
 * All use of this software is subject to the terms of the revised BSD
 * license.  You should have received a copy of this license along
 * with this source code in a file named "LICENSE."
 *
 * @file eaglView.mm
 * @author drose
 * @date 2009-04-10
 */

#import "EAGLView.h"

#import <QuartzCore/QuartzCore.h>
#import <OpenGLES/EAGLDrawable.h>

#include "pandabase.h"
#include "pnotify.h"
#include "iPhoneGraphicsWindow.h"

#define USE_DEPTH_BUFFER 1

// A class extension to declare private methods
@interface EAGLView ()

@property (nonatomic, retain) EAGLContext *context;

- (BOOL) createFramebuffer;
- (void) destroyFramebuffer;

@end


@implementation EAGLView

@synthesize context;


// You must implement this method
+ (Class)layerClass {
    return [CAEAGLLayer class];
}


- (id)initWithFrame:(CGRect)frame {
  if ((self = [super initWithFrame:frame])) {
    // Get the layer
    CAEAGLLayer *eaglLayer = (CAEAGLLayer *)self.layer;

    eaglLayer.opaque = YES;
    eaglLayer.drawableProperties = [NSDictionary dictionaryWithObjectsAndKeys:
                                                   [NSNumber numberWithBool:NO], kEAGLDrawablePropertyRetainedBacking, kEAGLColorFormatRGBA8, kEAGLDrawablePropertyColorFormat, nil];

    context = [[EAGLContext alloc] initWithAPI:kEAGLRenderingAPIOpenGLES1];

    if (!context || ![EAGLContext setCurrentContext:context]) {
      [self release];
      return nil;
    }
    _window = NULL;
    viewFramebuffer = 0;
    viewRenderbuffer = 0;
    depthRenderbuffer = 0;

    self.multipleTouchEnabled = YES;
  }
  return self;
}


- (void)presentView {
  [context presentRenderbuffer:GL_RENDERBUFFER_OES];
}

- (void)touchesBegan: (NSSet *)touches
           withEvent: (UIEvent *)event
{
  // Pass the multi-touch input to the _window for processing.
  _window->touches_began(touches, event);
  [super touchesBegan: touches withEvent: event];
}

- (void)touchesMoved: (NSSet *)touches
           withEvent: (UIEvent *)event
{
  _window->touches_moved(touches, event);
  [super touchesMoved: touches withEvent: event];
}

- (void)touchesEnded: (NSSet *)touches
           withEvent: (UIEvent *)event
{
  _window->touches_ended(touches, event);
  [super touchesEnded: touches withEvent: event];
}

- (void)touchesCancelled: (NSSet *)touches
               withEvent: (UIEvent *)event
{
  _window->touches_cancelled(touches, event);
  [super touchesCancelled: touches withEvent: event];
}

/*
- (void)touchesMoved:(NSSet *)touches withEvent:(UIEvent *)event;
- (void)touchesEnded:(NSSet *)touches withEvent:(UIEvent *)event;
- (void)touchesCancelled:(NSSet *)touches withEvent:(UIEvent *)event;
*/


- (void)layoutSubviews {
  [EAGLContext setCurrentContext:context];
  [self destroyFramebuffer];
  [self createFramebuffer];

  if (_window != NULL) {
    _window->rotate_window();
  }
}


- (BOOL)createFramebuffer {
  glGenFramebuffersOES(1, &viewFramebuffer);
  glGenRenderbuffersOES(1, &viewRenderbuffer);

  glBindFramebufferOES(GL_FRAMEBUFFER_OES, viewFramebuffer);
  glBindRenderbufferOES(GL_RENDERBUFFER_OES, viewRenderbuffer);
  [context renderbufferStorage:GL_RENDERBUFFER_OES fromDrawable:(CAEAGLLayer*)self.layer];
  glFramebufferRenderbufferOES(GL_FRAMEBUFFER_OES, GL_COLOR_ATTACHMENT0_OES, GL_RENDERBUFFER_OES, viewRenderbuffer);

  glGetRenderbufferParameterivOES(GL_RENDERBUFFER_OES, GL_RENDERBUFFER_WIDTH_OES, &backingWidth);
  glGetRenderbufferParameterivOES(GL_RENDERBUFFER_OES, GL_RENDERBUFFER_HEIGHT_OES, &backingHeight);

  if (USE_DEPTH_BUFFER) {
    glGenRenderbuffersOES(1, &depthRenderbuffer);
    glBindRenderbufferOES(GL_RENDERBUFFER_OES, depthRenderbuffer);
    glRenderbufferStorageOES(GL_RENDERBUFFER_OES, GL_DEPTH_COMPONENT16_OES, backingWidth, backingHeight);
    glFramebufferRenderbufferOES(GL_FRAMEBUFFER_OES, GL_DEPTH_ATTACHMENT_OES, GL_RENDERBUFFER_OES, depthRenderbuffer);
    glBindRenderbufferOES(GL_RENDERBUFFER_OES, viewRenderbuffer);
  }

  if(glCheckFramebufferStatusOES(GL_FRAMEBUFFER_OES) != GL_FRAMEBUFFER_COMPLETE_OES) {
    NSLog(@"failed to make complete framebuffer object %x", glCheckFramebufferStatusOES(GL_FRAMEBUFFER_OES));
    return NO;
  }

  // Make sure the buffer is initially cleared, so we don't look at whatever
  // happened to be in the framebuffer.
  glClearColor(0, 0, 0, 1);
  glClear(GL_COLOR_BUFFER_BIT);
  [context presentRenderbuffer:GL_RENDERBUFFER_OES];

  return YES;
}


- (void)destroyFramebuffer {
  if (viewFramebuffer != 0) {
    glDeleteFramebuffersOES(1, &viewFramebuffer);
    viewFramebuffer = 0;
  }

  if (viewRenderbuffer != 0) {
    glDeleteRenderbuffersOES(1, &viewRenderbuffer);
    viewRenderbuffer = 0;
  }

  if (depthRenderbuffer) {
    glDeleteRenderbuffersOES(1, &depthRenderbuffer);
    depthRenderbuffer = 0;
  }
}

- (void)dealloc {

    if ([EAGLContext currentContext] == context) {
        [EAGLContext setCurrentContext:nil];
    }

    [context release];
    [super dealloc];
}

@end
